#include "matrix.h"
#include "mylib.h"
#include <cassert>
#include <chrono>
#include <cuda_runtime.h> // cudaMalloc, cudaFree,cudaGetDeviceProperties
#include <cublas_v2.h>    //  new CUDA API, see https://docs.nvidia.com/cuda/cublas/index.html#new-and-legacy-cublas-api
#include <cmath>
#include <vector>
using namespace std;
using namespace std::chrono;    // timing

DenseMatrix::DenseMatrix(int const nrow, int const mcol)
    : _nrows(nrow), _mcols(mcol), _A(_nrows * _mcols)
{
    int const nm = max(_nrows, _mcols);
    auto const x = linspace(-5.0, 5.0, nm);

    vector<double> fx(nm);
    for (size_t k = 0; k < fx.size(); ++k) {
        fx[k] = sigmoid(-x[k]);
    }

    for (int i = 0; i < _nrows; ++i) {
        for (int j = 0; j < _mcols; ++j) {
            _A[i * _mcols + j] = fx[i] * fx[j];
        }
    }

}

DenseMatrix::DenseMatrix(int const nrow, int const mcol, double const val)
    : _nrows(nrow), _mcols(mcol), _A(_nrows * _mcols, val)
{
}


vector<double> DenseMatrix::Mult(vector<double> const &u) const
{
    assert( GetNcols() == static_cast<int>(u.size()) ); // #columns in matrix =? #elements in vector

    int const nrows = GetNrows();
    int const mcols = GetNcols();

    vector<double> f(nrows);                 // allocate resulting vector
    // matrix times vector: f := A * u
    for (int i = 0; i < nrows; ++i) {
        double tmp = 0.0;
        for (int j = 0; j < mcols; ++j) {
            tmp = tmp + _A[i * mcols + j] * u[j];
        }
        f[i] = tmp;
    }

    return f;
}

vector<double> DenseMatrix::MultT(vector<double> const &u) const
{
    assert( GetNrows() == static_cast<int>(u.size()) ); // #rows in matrix =? #elements in vector

    int const nrows = GetNrows();
    int const mcols = GetNcols();

    vector<double> f(mcols);                 // allocate resulting vector
    // matrix times vector: f := A^T * u
    for (int i = 0; i < mcols; ++i) {        // i and j loops are swapped in matrix access
        double tmp = 0.0;
        for (int j = 0; j < nrows; ++j) {
            tmp = tmp + _A[j * mcols + i] * u[j];
        }
        f[i] = tmp;
    }

    return f;
}

// loops j and i are swapped
vector<double> DenseMatrix::MultT_fast(vector<double> const &u) const
{
    assert( GetNrows() == static_cast<int>(u.size()) ); // #rows in matrix =? #elements in vector

    int const nrows = GetNrows();
    int const mcols = GetNcols();

    vector<double> f(mcols, 0.0);            // allocate resulting vector
    // matrix times vector: f := A^T * u
    for (int j = 0; j < nrows; ++j) {        // i and j loops are swapped in matrix access
        for (int i = 0; i < mcols; ++i) {
            f[i] += _A[j * mcols + i] * u[j];
        }
    }

    return f;
}


//     C = A*B
DenseMatrix DenseMatrix::Mult(DenseMatrix const &B) const
{
    DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?

    DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);
    int const acols = A.GetNcols();
    int const bcols = B.GetNcols();
    int const ccols = C.GetNcols();

    // GH ToDo: Three loops for matrix*matrix product
    for (int i = 0; i < C.GetNrows(); ++i) {
        for (int j = 0; j < C.GetNcols(); ++j) {
            double tmp = 0.0;
            for (int k = 0; k < A.GetNcols(); ++k) {
                tmp += A[i * acols + k] * B[k * bcols + j];
            }
            C[i * ccols + j] = tmp;
        }
    }

    return C;
}

//     C = A*B
DenseMatrix DenseMatrix::Mult_fast(DenseMatrix const &B) const
{
    DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?

    DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);         // fast
    int const acols = A.GetNcols();
    int const bcols = B.GetNcols();
    int const ccols = C.GetNcols();

    // GH ToDo: Three loops for matrix*matrix product
    for (int i = 0; i < C.GetNrows(); ++i) {
        for (int k = 0; k < A.GetNcols(); ++k) {            // fast
            // double tmp=0.0;                              // fast
            for (int j = 0; j < C.GetNcols(); ++j) {        // fast
                C[i * ccols + j] += A[i * acols + k] * B[k * bcols + j]; // fast
            }
            // C[i*ccols+j] = tmp;                          // fast
        }
    }

    return C;
}

DenseMatrix DenseMatrix::Mult_Blas(DenseMatrix const &B) const
{
//  C_(MxN) := A_(MxK) * B(KxN)
//  rowise storage in all three matrices

    DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?
    DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);

    double const *const pA = A.data();
    double const *const pB = B.data();
    double       *const pC = C.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-tutorial-c-multiplying-matrices-using-dgemm
//      http://www.netlib.org/lapack/explore-html/d1/d54/group__double__blas__level3_gaeda3cbd99c8fb834a60a6412878226e1.html
//      http://www.netlib.org/lapack/explore-html/dc/d18/cblas__dgemm_8c.html
    unsigned int const M = C.GetNrows();
    unsigned int const N = C.GetNcols();
    unsigned int const K = A.GetNcols();
    unsigned int const LDA = K;
    unsigned int const LDB = N;
    unsigned int const LDC = N;

    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                M, N, K,
                alpha, pA, LDA,
                pB, LDB,
                beta, pC, LDC);

    return C;
}

////  C_(MxN) := A_(MxK) * B(KxN)
////  rowise storage in all three matrices
////  BUT cUBLAS expects columnwise storage
//DenseMatrix DenseMatrix::Mult_cuBlas_row(DenseMatrix const &B) const
//{
    //DenseMatrix const  &A = *this;
    //assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?
    //DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);

    //double const *const pA_h = A.data();    // host data
    //double const *const pB_h = B.data();
    //double       *const pC_h = C.data();

    //double const alpha = 1.0;
    //double const beta = 0.0;
////      https://software.intel.com/en-us/mkl-tutorial-c-multiplying-matrices-using-dgemm
////      http://www.netlib.org/lapack/explore-html/dc/d18/cblas__dgemm_8c.html
    //unsigned int const M = C.GetNrows();
    //unsigned int const N = C.GetNcols();
    //unsigned int const K = A.GetNcols();
    //unsigned int const LDA = K;
    //unsigned int const LDB = N;
    //unsigned int const LDC = N;

////  -- Initialize cuBLAS, check GPU
////  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
    //cudaError_t  cudaStat;               //  cudaMalloc  status
    //cublasStatus_t  stat;                //  CUBLAS  functions  status
    //cublasHandle_t  handle;

    //stat = cublasCreate (& handle );   //  initialize  CUBLAS  context
    //if (stat != CUBLAS_STATUS_SUCCESS) {
        //fprintf(stderr, "ERROR: cublasInit() failed!\n");
        //exit(1);
    //}

    //cudaDeviceProp prop;
    //cudaGetDeviceProperties (&prop, 0);
    //cout << endl << "We work on a " << prop.name << " GPU  with " <<
         //prop.multiProcessorCount << " Multiprocessors" << endl;

////  -- Move data from host to device
//// allocate memory on device
    //double *A_d, *B_d, *C_d; // device data
    //cudaStat=cudaMalloc((void **) &A_d, M*K * sizeof(*A_d));
    //cudaStat=cudaMalloc((void **) &B_d, K*N * sizeof(*B_d));
    //cudaStat=cudaMalloc((void **) &C_d, M*N * sizeof(*C_d));

//// copy data:  host --> device
    //cublasSetMatrix(M, K, sizeof(*A_d), pA_h, M, A_d, M);
    //cublasSetMatrix(K, N, sizeof(*B_d), pB_h, K, B_d, K);
    //cublasSetMatrix(M, N, sizeof(*C_d), pC_h, M, C_d, M);

////  arithm.
////  CUBLAS_OP_T , see https://docs.nvidia.com/cuda/cublas/index.html#cublasoperation_t
    //stat=cublasDgemm(handle,CUBLAS_OP_N,CUBLAS_OP_N,M,N,K,&alpha,A_d,M,B_d,K,&beta,C_d,M);
    ////stat=cublasDgemm(handle,CUBLAS_OP_T,CUBLAS_OP_T,M,N,K,&alpha,A_d,K,B_d,N,&beta,C_d,N);
    //if (stat != CUBLAS_STATUS_SUCCESS) {
        //fprintf(stderr, "cublasDgemm ERROR: %s\n",stat);
        //exit(1);
    //}

////  copy data: device --> host
    //stat=cublasGetMatrix(M, N, sizeof(*C_d),C_d, M, pC_h, M);

////  -- Free all cuBlas ressources
    //cudaFree(C_d);
    //cudaFree(B_d);
    //cudaFree(A_d);
    //cublasDestroy(handle);

    //return C;
//}


//  C_(MxN) := A_(MxK) * B(KxN)
//  rowise storage in all three matrices
//  BUT cuBLAS expects columnwise storage
//  see:
//  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
DenseMatrix DenseMatrix::Mult_cuBlas(DenseMatrix const &B) const
{
    DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?
    DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);

    double const *const pA_h = A.data();    // host data
    double const *const pB_h = B.data();
    double       *const pC_h = C.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-tutorial-c-multiplying-matrices-using-dgemm
//      http://www.netlib.org/lapack/explore-html/dc/d18/cblas__dgemm_8c.html
    unsigned int const M = C.GetNrows();
    unsigned int const N = C.GetNcols();
    unsigned int const K = A.GetNcols();
    unsigned int const LDA = K;
    unsigned int const LDB = N;
    unsigned int const LDC = N;

//  -- Initialize cuBLAS, check GPU
//  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
    cudaError_t  cudaStat;               //  cudaMalloc  status
    cublasStatus_t  stat;                //  CUBLAS  functions  status
    cublasHandle_t  handle;

    stat = cublasCreate (& handle );   //  initialize  CUBLAS  context
    if (stat != CUBLAS_STATUS_SUCCESS) {
        fprintf(stderr, "ERROR: cublasInit() failed!\n");
        exit(1);
    }

    //cudaDeviceProp prop;
    //cudaGetDeviceProperties (&prop, 0);
    //cout << endl << "We work on a " << prop.name << " GPU  with " <<
         //prop.multiProcessorCount << " Multiprocessors" << endl;

//  -- Move data from host to device
// allocate memory on device
    double *A_d, *B_d, *C_d; // device data
    cudaStat=cudaMalloc((void **) &A_d, M*K * sizeof(*A_d));
    cudaStat=cudaMalloc((void **) &B_d, K*N * sizeof(*B_d));
    cudaStat=cudaMalloc((void **) &C_d, M*N * sizeof(*C_d));

//  copy data:  host --> device
//  Interprete our rowise storage on host as columwise storage on device
    stat=cublasSetMatrix(K, M, sizeof(*A_d), pA_h, K, A_d, K);       // A^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "A: cublasSetMatrix ERROR: %s\n",stat); exit(1);}
    stat=cublasSetMatrix(N, K, sizeof(*B_d), pB_h, N, B_d, N);       // B^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "B: cublasSetMatrix ERROR: %s\n",stat); exit(1);}
    stat=cublasSetMatrix(N, M, sizeof(*C_d), pC_h, N, C_d, N);      // C^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "C: cublasSetMatrix ERROR: %s\n",stat); exit(1);}

//for (int k=0; k<100; ++k)
{
//  arithm.
//  CUBLAS_OP_N , see https://docs.nvidia.com/cuda/cublas/index.html#cublasoperation_t
    // C = A*B   // only correct if we use also columnwise storage
    //stat=cublasDgemm(handle,CUBLAS_OP_N,CUBLAS_OP_N,M,N,K,&alpha,A_d,M,B_d,K,&beta,C_d,M);
    // C^T = B^T * A^T
    stat=cublasDgemm(handle,CUBLAS_OP_N,CUBLAS_OP_N,N,M,K,&alpha,B_d,N,A_d,K,&beta,C_d,N);
            //if (stat != CUBLAS_STATUS_SUCCESS) { fprintf(stderr, "cublasDgemm ERROR: %s\n",stat); exit(1);}
}
//  copy data: device --> host
//  Interprete columwise storage on device as rowise storage on host
    stat=cublasGetMatrix(N,M, sizeof(*C_d),C_d, N, pC_h, N);        // C^T ==> C on host
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "C: cublasGetMatrix ERROR: %s\n",stat); exit(1);}

//  -- Free all cuBlas ressources
    cudaFree(C_d);
    cudaFree(B_d);
    cudaFree(A_d);
    cublasDestroy(handle);

    return C;
}

DenseMatrix DenseMatrix::Mult_cuBlas_Test(DenseMatrix const &B, int const loopcount, double &seconds) const
{
    DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?
    DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);

    double const *const pA_h = A.data();    // host data
    double const *const pB_h = B.data();
    double       *const pC_h = C.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-tutorial-c-multiplying-matrices-using-dgemm
//      http://www.netlib.org/lapack/explore-html/dc/d18/cblas__dgemm_8c.html
    unsigned int const M = C.GetNrows();
    unsigned int const N = C.GetNcols();
    unsigned int const K = A.GetNcols();
    unsigned int const LDA = K;
    unsigned int const LDB = N;
    unsigned int const LDC = N;

//  -- Initialize cuBLAS, check GPU
//  https://developer.nvidia.com/sites/default/files/akamai/cuda/files/Misc/mygpu.pdf
    cudaError_t  cudaStat;               //  cudaMalloc  status
    cublasStatus_t  stat;                //  CUBLAS  functions  status
    cublasHandle_t  handle;

    stat = cublasCreate (& handle );   //  initialize  CUBLAS  context
    if (stat != CUBLAS_STATUS_SUCCESS) {
        fprintf(stderr, "ERROR: cublasInit() failed!\n");
        exit(1);
    }

    cudaDeviceProp prop;
    cudaGetDeviceProperties (&prop, 0);
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
         prop.multiProcessorCount << " Multiprocessors" << endl;

//  -- Move data from host to device
// allocate memory on device
    double *A_d, *B_d, *C_d; // device data
    cudaStat=cudaMalloc((void **) &A_d, M*K * sizeof(*A_d));
    cudaStat=cudaMalloc((void **) &B_d, K*N * sizeof(*B_d));
    cudaStat=cudaMalloc((void **) &C_d, M*N * sizeof(*C_d));

//  copy data:  host --> device
//  Interprete our rowise storage on host as columwise storage on device
    stat=cublasSetMatrix(K, M, sizeof(*A_d), pA_h, K, A_d, K);       // A^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "A: cublasSetMatrix ERROR: %s\n",stat); exit(1);}
    stat=cublasSetMatrix(N, K, sizeof(*B_d), pB_h, N, B_d, N);       // B^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "B: cublasSetMatrix ERROR: %s\n",stat); exit(1);}
    stat=cublasSetMatrix(N, M, sizeof(*C_d), pC_h, N, C_d, N);      // C^T  on device
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "C: cublasSetMatrix ERROR: %s\n",stat); exit(1);}

    auto t1 = system_clock::now();
    for (int k=0; k<loopcount; ++k)
    {
//  arithm.
//  CUBLAS_OP_N , see https://docs.nvidia.com/cuda/cublas/index.html#cublasoperation_t
    // C = A*B   // only correct if we use also columnwise storage
    //stat=cublasDgemm(handle,CUBLAS_OP_N,CUBLAS_OP_N,M,N,K,&alpha,A_d,M,B_d,K,&beta,C_d,M);
    // C^T = B^T * A^T
        cudaDeviceSynchronize();
        stat=cublasDgemm(handle,CUBLAS_OP_N,CUBLAS_OP_N,N,M,K,&alpha,B_d,N,A_d,K,&beta,C_d,N);
           //if (stat != CUBLAS_STATUS_SUCCESS) { fprintf(stderr, "cublasDgemm ERROR: %s\n",stat); exit(1);}
    }
    cudaDeviceSynchronize();
    auto t2 = system_clock::now();
    auto scd=duration_cast<microseconds>(t2 - t1)/loopcount; //1e6/
    //seconds = static_cast<double>(scd);
    //  https://stackoverflow.com/questions/57538507/how-to-convert-stdchronoduration-to-double-seconds
    seconds = std::chrono::duration<double>(scd).count();

//  copy data: device --> host
//  Interprete columwise storage on device as rowise storage on host
    stat=cublasGetMatrix(N,M, sizeof(*C_d),C_d, N, pC_h, N);        // C^T ==> C on host
            //if (stat != CUBLAS_STATUS_SUCCESS) {fprintf(stderr, "C: cublasGetMatrix ERROR: %s\n",stat); exit(1);}

//  -- Free all cuBlas ressources
    cudaFree(C_d);
    cudaFree(B_d);
    cudaFree(A_d);
    cublasDestroy(handle);

    return C;
}




bool DenseMatrix::operator==(DenseMatrix const &B) const
{
    DenseMatrix const  &A = *this;
    assert(A.GetNrows() == B.GetNrows());
    assert(A.GetNcols() == B.GetNcols());

    double const eps = 1e-6;
    bool bsame = true;
    for (int k = 0; k < A.GetNrows()*A.GetNcols(); ++k) {
        bool bk = fabs(A[k] - B[k]) <= eps * (1 + 0.5 * ( fabs(A[k]) + fabs(B[k]) ));
        if (!bk) {
            cerr << "matrices differ at [" << k / A.GetNcols() << "," << k % A.GetNcols() << "]";
            cerr << " :  " << A[k] << "  vs. " << B[k] << endl;
        }
        bsame &= bk;
    }
    return bsame;
}



//------------  Blas_DenseMatrix  ---------------------------------------------------

Blas_DenseMatrix::Blas_DenseMatrix(int const nrow, int const mcol)
    : DenseMatrix(nrow, mcol)
{}

Blas_DenseMatrix::Blas_DenseMatrix(int const nrow, int const mcol, double const val)
    : DenseMatrix(nrow, mcol, val)
{}


vector<double> Blas_DenseMatrix::Mult(vector<double> const &u) const
{
    assert( GetNcols() == static_cast<int>(u.size()) ); // #columns in matrix =? #elements in vector

    int const nrows = GetNrows();
    vector<double> f(nrows);                 // allocate resulting vector

    double const *const pA = _A.data();
    double const *const pu = u.data();
    double *const       pf = f.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-developer-reference-c-cblas-gemv
    unsigned int const M = GetNrows();
    unsigned int const N = GetNcols();
    unsigned int const LDA = N;

    cblas_dgemv(CblasRowMajor,
                CblasNoTrans,
                M, N,
                alpha, pA, LDA,
                pu, 1,
                beta, pf, 1);
    return f;
}

vector<double> Blas_DenseMatrix::MultT(vector<double> const &u) const
{
    assert( GetNrows() == static_cast<int>(u.size()) ); // #rows in matrix =? #elements in vector

    int const mcols = GetNcols();
    vector<double> f(mcols);                 // allocate resulting vector
    // matrix times vector: f := A^T * u
    double const *const pA = _A.data();
    double const *const pu = u.data();
    double *const       pf = f.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-developer-reference-c-cblas-gemv
    unsigned int const M = GetNrows();
    unsigned int const N = GetNcols();
    unsigned int const LDA = M;

    cblas_dgemv(CblasRowMajor, CblasTrans,
                M, N,
                alpha, pA, LDA,
                pu, 1,
                beta, pf, 1);

    return f;
}

/** 	Matrix-matrix multiplication (rowise) access.
 *
 * 	@param[in] B  matrix
 * 	@return resulting matrix @f$ A*B @f$
 *
 */
Blas_DenseMatrix Blas_DenseMatrix::Mult(Blas_DenseMatrix const &B) const
{
//  C_(MxN) := A_(MxK) * B(KxN)
//  rowise storage in all three matrices

    Blas_DenseMatrix const  &A = *this;
    assert( A.GetNcols() == B.GetNrows() ); // inner dimensions equal?
    Blas_DenseMatrix C(A.GetNrows(), B.GetNcols(), 0.0);

    double const *const pA = A.data();
    double const *const pB = B.data();
    double *const pC = C.data();

    double const alpha = 1.0;
    double const beta = 0.0;
//      https://software.intel.com/en-us/mkl-tutorial-c-multiplying-matrices-using-dgemm
//      http://www.netlib.org/lapack/explore-html/d1/d54/group__double__blas__level3_gaeda3cbd99c8fb834a60a6412878226e1.html
//      http://www.netlib.org/lapack/explore-html/dc/d18/cblas__dgemm_8c.html
    unsigned int const M = C.GetNrows();
    unsigned int const N = C.GetNcols();
    unsigned int const K = A.GetNcols();
    unsigned int const LDA = K;
    unsigned int const LDB = N;
    unsigned int const LDC = N;

    cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
                M, N, K,
                alpha, pA, LDA,
                pB, LDB,
                beta, pC, LDC);

    return C;
}


//----------------------------------------------------------------------

//vector<double> linspace(double const a, double const b, int const n)
//{
    //vector<double> x(n);
    //double const h = (b - a) / (n - 1.0);
    //for (size_t k = 0; k < x.size(); ++k) {
        //x[k] = k * h + a;
    //}
    //return x;
//}
//----------------------------------------------------------------------
TensorMatrix::TensorMatrix(vector<double> const &ucol, vector<double> const &vrow)
    : _nrows(ucol.size()), _mcols(vrow.size()), _ucol(ucol), _vrow(vrow)
{
}


vector<double> TensorMatrix::Mult(vector<double> const &x) const
{
    assert( _vrow.size() == x.size() );      // #columns in matrix =? #elements in vector

    int const nrows = GetNrows();
    int const mcols = GetNcols();

    vector<double> f(nrows);                 // allocate resulting vector

    double s = 0.0;
    for (int j = 0; j < mcols; ++j) {        // inner product
        s += _vrow[j] * x[j];
    }

    for (int i = 0; i < nrows; ++i) {        // scale a vector
        f[i] = s * _ucol[i];
    }

    return f;
}

vector<double> TensorMatrix::MultT(vector<double> const &y) const
{
    assert( _ucol.size() == y.size() );      // #rows in matrix =? #elements in vector

    int const nrows = GetNrows();
    int const mcols = GetNcols();

    vector<double> f(mcols);                 // allocate resulting vector

    double s = 0.0;
    for (int i = 0; i < nrows; ++i) {        // inner product
        s += _ucol[i] * y[i];
    }

    for (int j = 0; j < mcols; ++j) {        // scale a vector
        f[j] = s * _vrow[j];
    }

    return f;
}
