// originates from Ruetsch/Oster: Getting Started with CUDA
#include <algorithm>
#include <cassert>
#include <cmath>
#include <iostream>

__global__ void inc_gpu(float *const a, int N);
__global__ void exp_gpu(float *const b, float const *const a, int N);
__global__ void log_gpu(float *const b, float const *const a, int N);

__global__ void inc_gpu(float *const a, int N)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < N)
        a[idx] = a[idx] + 1;
}

__global__ void exp_gpu(float *const b, float const *const a, int N)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < N)
        b[idx] = std::exp(a[idx]);
}

__global__ void log_gpu(float *const b, float const *const a, int N)
{
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < N)
        b[idx] = std::log(a[idx]);
}


using namespace std;

int main(void)
{
    int const N = 1400;
    int const nBytes = N * sizeof(float);

    int const blockSize = 64;
    int const numBlocks = (N + blockSize - 1) / blockSize;

    float *a_h = new float [nBytes];     // host data
    float *b_h = new float [nBytes];     // host data
    float *a_d, *b_d;                    // device data
    cudaMalloc((void **) &a_d, nBytes);
    cudaMalloc((void **) &b_d, nBytes);

    for (int i = 0; i < N; i++) a_h[i] = 100.0f + static_cast<float>(i);

    cudaMemcpy(a_d, a_h, nBytes, cudaMemcpyHostToDevice);    //  a_d <- a_h
    cudaMemcpy(b_d, a_d, nBytes, cudaMemcpyDeviceToDevice);  //  b_d <- a_d

// ---------------------------------------------------------

// Manipulate on GPU
    inc_gpu <<< numBlocks, blockSize>>>(b_d, N);             //  b_d := b_d+1.0

    cudaMemcpy(b_h, b_d, nBytes, cudaMemcpyDeviceToHost);    //  b_h <- b_d

// Check on CPU
    for (int i = 0; i < N; i++) {assert( a_h[i] == b_h[i] - 1.0f );}
    cout << "Check 1  OK" << endl;
// ---------------------------------------------------------

    float *c_h = new float [nBytes];     // host data
    float *c_d;                          // device data
    cudaMalloc((void **) &c_d, nBytes);


// Manipulate on GPU
    log_gpu <<< numBlocks, blockSize>>>(b_d,a_d, N);            //  b_d := exp(a_d)
    exp_gpu <<< numBlocks, blockSize>>>(c_d,b_d, N);            //  c_d := log(b_d)

    cudaMemcpy(c_h, c_d, nBytes, cudaMemcpyDeviceToHost);    //  c_h <- c_d

// Check on CPU
    //for (int i = 0; i < N; i++) assert( a_h[i] == c_h[i] );
    float const EPS=1e-5f;                                    // relatve accuracy
    bool bb=equal(a_h, a_h+N, c_h,
          [EPS] (float x, float y) -> bool
          {
			bool const ba = abs(x-y)<=EPS*(abs(x)+abs(y))/2;
			if (!ba) cout << x << "  " << y << "  " << x-y << endl;
			return ba;
		   }
    );
    assert(bb);
    cout << "Check 2  OK" << endl;

// ---------------------------------------------------------
    delete [] c_h;
    delete [] b_h;
    delete [] a_h;
    cudaFree(a_d);
    cudaFree(b_d);
    cudaFree(c_d);

    return 0;
}
