// scalar product: 3: recursive local summation and summation on GPU
// Profiler:  export CUDA_PROFILE=1
//   Attention:   vector length ==  k*blockDim.x
#include <cassert>
#include <iostream>
#include <ctime>
using namespace std;

//    kernel function on device
/** 	Device function. Calculates the inner product per thread block
	@param[out] sg	vector on device storing the partial inner products
	@param[in] x	vector on device
	@param[in] y	vector on device
	@param[in] N	number of vector elements
    @warning shared memory of blocksize * sizeof(float) bytes has to be required on kernel call
*/
__global__ void scalar(float *sg, const float *x, const float *y, int N)
{
    extern __shared__ float sdata[];
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int str = gridDim.x * blockDim.x;

    float ss = 0.0;
    for (int i = idx; i < N; i += str)
        ss += x[i] * y[i];
    sdata[threadIdx.x] = ss;

    __syncthreads();
    for (int s = blockDim.x >> 1; s > 0; s >>= 1) {
        if (threadIdx.x < s) {
            sdata[threadIdx.x] += sdata[threadIdx.x + s];
        }
        __syncthreads();
    }
    if (threadIdx.x == 0)  sg[blockIdx.x] = sdata[0];
}

/** 	Device function. Kernel function for addition on  o n e  block on device
	@param[inout] s	vector on device , its sum is stored in @p s[0]
	@param[in] N	number of vector elements
    @warning shared memory has to be allocated
    @see scalar
*/
__global__ void add_1_block(float *s, int N)
{
    if (N > blockDim.x) return;
    extern __shared__ float sdata[];
    const int tid = threadIdx.x;
    if (tid < N) sdata[tid] = s[tid];
    else       sdata[tid] = 0.0;

    __syncthreads();
    for (int s = blockDim.x / 2; s > 0; s >>= 1) {
        if (tid < s) {
            sdata[tid] += sdata[tid + s];
        }
        __syncthreads();
    }
    if (tid == 0)  s[0] = sdata[0];
}


//----------------------------------------------------------------------
static bool b_init=false;

/** 	Host function. Inner product calculated with device vectors.
 *  The implementation consists of two kernel with a temporary array for data transfer between them.
 *  The necessary dynamic memory alocation is done only in the first function call (global variable @p b_init).
 *  This  dynamic memory is not deallocated afterwards.
	@param[in] N	number of vector elements
	@param[in] x_d	vector on device
	@param[in] y_d	vector on device
	@return 	resulting Euclidian inner product <x_d,y_d> on host
	@warning    The dynamic memory is not deallocated afterwards.
*/
float dscapr_GPU(const int N, const float x_d[], const float y_d[])
{
	static float *s_d;
	static dim3 dimBlock, dimGrid;
	static int sBytes;
	if (!b_init)   // only  in the first function call
	{
		b_init = !b_init;
		cudaDeviceProp prop;
        cudaGetDeviceProperties (&prop, 0);

        const int blocksize = 8 * 64,
                  gridsize = prop.multiProcessorCount;
        sBytes   = blocksize * sizeof(float);
        dimBlock = dim3(blocksize);
        dimGrid  = dim3(4*gridsize);

        //float s_d;
        cudaMalloc((void **) &s_d, blocksize * sizeof(float)); // temp. memory on device
        cout << endl << gridsize << " x " << blocksize << " Threads\n";
        cout << "\n###  Fast version of inner product on device is used. ###\n\n";
    }

// call the kernel function with  dimGrid.x * dimBlock.x threads
    scalar <<< dimGrid, dimBlock, sBytes>>>(s_d, x_d, y_d, N);
//    power of 2 as number of threads per block
    const unsigned int oneBlock = 2 << (int)ceil(log(dimGrid.x + 0.0) / log(2.0));
    add_1_block <<< 1, oneBlock, oneBlock *sizeof(float)>>>(s_d, dimGrid.x);
// copy data:  device --> host
    float sg;
    cudaMemcpy(&sg, s_d, sizeof(float), cudaMemcpyDeviceToHost);

    //cudaFree(s_d);                   // GH: ToDo: Free memory after the end of all computations
    return sg;
}

//----------------------------------------------------------------------

//  main function on host
int main(void)
{
    float *x_h, *y_h; // host data
    float *x_d, *y_d; // device data
    float sum, tstart, tgpu;
    const int N = 100000000,
    //const int N = 14000000,
              nBytes = N * sizeof(float);
    const int LOOP = 100;


    // https://devblogs.nvidia.com/how-query-device-properties-and-handle-errors-cuda-cc/
    int nDevices;
    cudaError_t err = cudaGetDeviceCount(&nDevices);
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));   
    
    cudaDeviceProp prop;
    // https://www.cs.cmu.edu/afs/cs/academic/class/15668-s11/www/cuda-doc/html/group__CUDART__DEVICE_g5aa4f47938af8276f08074d09b7d520c.html
    err = cudaGetDeviceProperties (&prop, 0); 
    if (err != cudaSuccess) printf("%i : %s\n", err, cudaGetErrorString(err));   
    cout << endl << "We work on a " << prop.name << " GPU  with " <<
    prop.multiProcessorCount << " Multiprocessors" << endl;


    x_h = new float [N];
    y_h = new float [N];
// allocate memory on device
    cudaMalloc((void **) &x_d, nBytes);
    cudaMalloc((void **) &y_d, nBytes);
//cudaMalloc((void **) &s_d, resBytes);

    float ss = 0.0;
    for (int i = 0; i < N; i++) {
        x_h[i] = (i % 137) + 1;
        y_h[i] = 1.0 / x_h[i];
        ss += x_h[i] * y_h[i];
    }

      cout << "\nN = " << N << endl;


// copy data:  host --> device
    cudaMemcpy(x_d, x_h, nBytes, cudaMemcpyHostToDevice);
    cudaMemcpy(y_d, y_h, nBytes, cudaMemcpyHostToDevice);
    
    sum = dscapr_GPU(N, x_d, y_d); // GH: Dummy call to set up memory

    cudaDeviceSynchronize();
    tstart = clock();
    for (int k = 0; k < LOOP; ++k) {
        sum = dscapr_GPU(N, x_d, y_d);
    }
    cudaDeviceSynchronize();
    tgpu = (clock() - tstart) / CLOCKS_PER_SEC / LOOP;

    cout << "GPU: " << tgpu << " sec.; " <<
         2e-9 * N / tgpu << " GFLOPS;  " <<
         2.0 * N *sizeof(float) / tgpu / 1024 / 1024 / 1024 << " GiB/sec.\n ";

    if (static_cast<int>(sum) != N) {
        cout << " !!  W R O N G  result  " << endl;
        cout << sum << "  vs.  " << N << "(" << ss << ")" << endl;
    }
    else
        cout << " correct  result  " << sum << endl;

    delete [] x_h;
    delete [] y_h;
    cudaFree(x_d);
    cudaFree(y_d); //cudaFree(s_d);
    return 0;
}
